#! /usr/bin/perl

my $location = substr( $0, 0, length( $0 ) - 3 );

my $notConnectedPattern = qr/failed to get device iterator/im;
my $linkModeFailurePattern = qr/failed to read GP32 information from GP32/im;
my $failedToRemovePattern = qr/Error: failed to remove/im;
my $failedToCopyPattern = qr/Error: failed to copy file/im;
my $failedToReadDirectories = qr/Error: failed to read directories from SMC/im;
my $failedToSetBasepath = qr/Error: no basepath defined/im;

my $statusPattern = qr/Total (\d*) entries and (\d*) bytes/;
my $smcIDPattern = qr/SMC Unique ID:\s*(\w*)/m;

my $smcID = "";

$output = &DoCommand( "info -T1" );
if( $output =~ /$smcIDPattern/ )
{
	$smcID = $1;
	&Log( "SMC ID:".$smcID );
}
else
{
	&Log( "Failed to locate SMC ID" );
}

if( $ARGV[ 0 ] eq "clear" && $smcID ne "" )
{
	&Log( "Clearing cache ".$smcID );
	unlink $location.".".$smcID.".cache"; 
}

$output = &DoCommand( "end" ) if( $ARGV[ 0 ] eq "end" );
$output = &DoCommand( "rm $ARGV[ 1 ]" ) if( $ARGV[ 0 ] eq "rm" );
$output = &DoCommand( "ul $ARGV[ 1 ] \'$ARGV[ 2 ]\'" ) if( $ARGV[ 0 ] eq "ul" );
$output = &DoCommand( "mkdir $ARGV[ 1 ]" ) if( $ARGV[ 0 ] eq "mkdir" );
$output = &DoCommand( "format" ) if( $ARGV[ 0 ] eq "format" );

if( $ARGV[ 0 ] eq "ls" ) 
{
	if( $ARGV[ 1 ] eq "")
	{
		&ReadCache if( $smcID ne "" ); # Check for cache first...
		if( $cache ne "" )
		{
			$output = $cache;
		}
		else
		{
			&Log( "Read directory from GP32..." );
			$output = &ReadFolders if ($ARGV[ 0 ] eq "ls" );
			$errorStatus = &CheckStatus( $output );
			&WriteCache( $output ) if( $errorStatus eq "OK" );
		}
	}
	else
	{
		$output = &ReadFolder( $ARGV[ 1 ], 0 );
	}
}

if( $smcID ne "" && $ARGV[ 0 ] eq "refresh" )
{
	&ReadCache;
	$output = &RefreshCachedFolder( $ARGV[ 1 ], $cache ) 
}

$errorStatus = &CheckStatus( $output );

if( $errorStatus eq "OK" )
{
	&Log( "OK" );
	print $output;
}
else
{
	&Log( $errorStatus );
	print $errorStatus;
}

exit;

sub ReadCache
{
	if( -e $location.".".$smcID.".cache" )
	{
		&Log( "Reading directory from cache ".$smcID."..." );
		open( FILE, $location.".".$smcID.".cache" );
		@lines = <FILE>;
		close( FILE );
		$cache = join( "", @lines );
	}
}

sub WriteCache
{
	my $data = shift;
	
	if( $smcID ne "" )
	{
		&Log( "Writing cache ".$smcID."..." );
		open( FILE, ">".$location.".".$smcID.".cache" ); # Cache output of ls...
		print FILE $data;		
		close( FILE );
	}
}

sub ReadFolder
{
#	return ">/GPMM/ C64:7:487351\n+PARADROI.D64:170 Kb\n+BCSQUEST.D64:170 Kb\n+TRAPDOOR.PRG:42 Kb\n+THRUST.T64:22 Kb\n+CHUCKIE2.PRG:34 Kb\n+URIDIUM.PRG:34 Kb\n>/GPMM/C64/ FRODO:2:576\n+TMP0:288\n+TMP1:288\n</GPMM/C64/FRODO/\n</GPMM/C64/\n";

	my $name = shift;
	my $recurse = shift;
	
	$name .= "/" if( substr( $name, -1 ) ne "/" );
	
	my $rawOutput = &DoCommand( "ls -T2 -l $name" );

	$errorStatus = &CheckStatus( $rawOutput );
	return $errorStatus if( $errorStatus ne "OK" );
	
	my @lines = split( /\n/, $rawOutput );
	@lines = @lines[ 2 .. $#lines ]; # Strip off Mr. Spiv's Header

	my $contents = "";

	my $status = @lines[ 0 ];

	if( $status =~ /$statusPattern/ ) # Extract stats
	{
		my @parts = split( /\//, $name );
		my $path = join( "/", @parts[ 0 .. $#parts - 1 ] );
		my $folderName = @parts[ $#parts ];
#		$path = "/" if( $path eq "" );
		$contents .= ">".$folderName.":".$1.":".$2."\n"  if( $folderName ne "" );
	}
	else
	{
		return $rawOutput;
	}
	
	foreach my $item (@lines[ 1 .. $#lines ])
	{
		$item =~ s/^\s+//;	
		$item =~ s/\s+$//;
		$item =~ s/\s+/ /g;
		my @fields = split( / /, $item );
		my $filename = @fields[ 5 ];
		my $size = @fields[ 0 ];
		if( $size > 1024*1024 ) { $size = int( $size / (1024*1024) )." Mb"; }
		elsif( $size > 1024 ) { $size = int( $size / 1024 )." Kb"; }
		if( substr( $filename, -1 ) eq "/" )
		{
			if( $recurse == 1 )
			{
				$contents .= &ReadFolder( $name.$filename, $recurse );
			}
		}	
		else
		{
			$contents .= "+".$filename.":".$size."\n";
		}
#		$contents .= substr( $filename, -1 ) eq "/" ? &ReadFolder( $name.$filename ) : "+".$filename.":".$size."\n";
	}
	
	chop( $name ) if( substr( $name, -1 ) eq "/" );
	$contents .= "<".$name."\n";
	
	return $contents;
}
sub ReadFolders
{
	return &ReadFolder( "/", 1 );
}

sub Log
{
	my $text = shift;
	
	open( FILE, ">>".$location.".log" );
	print FILE $text."\n";
	close( FILE );
}

sub DoCommand
{
	my $command = shift;

	&Log( $command );	
	return `$location $command 2>&1`;
}

sub CheckStatus
{
	my $data = shift;
	return "ERROR: Failed to copy ".$ARGV[1] if( $output =~ /$failedToCopyPattern/ );
	return "ERROR: Failed to delete ".$ARGV[1] if( $output =~ /$failedToRemovePattern/ );
	return "ERROR: Failed to talk to GP32 - is link mode enabled?" if( $output =~ /$linkModeFailurePattern/ );
	return "ERROR: GP32 not detected - is it connected?" if( $output =~ /$notConnectedPattern/ );	
	return "ERROR: FAILED TO READ" if( $output =~ /$failedToReadDirectories/ );
	return "ERROR: No basepath defined" if( $output =~ /$failedToSetBasepath/ );
	return "OK";
}

sub RefreshCachedFolder
{
	my $folder = shift;
	my $contents = shift;
		
	&Log( "Refreshing cache of folder ".$folder );
		
	&ReadCache;
	
	$newContents = &ReadFolder( $folder, 1 );

	my @lines = split( /\n/, $contents );
	my @newLines = split( /\n/, $newContents );

	my $path   = "/";
	my $offset = 0;
	my $start  = -1;
	foreach $line (@lines)
	{
		my $first = substr( $line, 0, 1 );
		if( $first eq ">" )
		{
			my( $name, $entries, $size ) = split( /:/, substr( $line, 1 ) );
			$path .= $name."/";
			$start = $offset if( $path eq $folder );
		}
		elsif( $first eq "<" )
		{
			my @fields = split( /\//, substr( $line, 1 ) );
			$path = join( "/", @fields[ 0 .. $#fields - 1 ] );
			$path = "/" if( $path eq "" );
			my $name = substr( $line, 1, length( $line ) - 1 );
			$name .= "/" if( substr( $line, -1 ) ne "/" );
			if( $name eq $folder )
			{
				$end = $offset;
				last;
			}
		}
		$offset++;
	}

	if( $start > -1 && $end > $start )
	{
		if( $folder ne "/" && $folder ne "" )
		{
			$start++;
			$end--;
		}
		splice( @lines, $start, $end - $start + 1, @newLines );
		$contents = join( "\n", @lines );
		&WriteCache( $contents );
		return join("\n", @lines[ $start..$end ] );
	}
	else
	{
		return $contents;
	}
}

sub AddToFolder
{
	my $folder = shift;
	my $entry = shift;
	
	my $offset = 0;
	foreach $line (@lines)
	{
		if( substr( $line, 0, 1 ) eq ">" )
		{
			my( $parent, $name ) = split( / /, substr( $line, 1 ) );
			my( $name, $entries, $size ) = split( /:/, $name );
			$parent = "" if( $parent eq "/" );
			if( $parent."/".$name eq $folder )
			{
				splice( @lines, $offset, 0, $entry."\n" );
				last;
			}
		}
		$offset++;
	}
	print "@lines";
}